package eu.eidas.auth.engine;

import java.io.IOException;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.collect.ImmutableMap;

import eu.eidas.auth.commons.EidasErrorKey;
import eu.eidas.auth.commons.EidasErrors;
import eu.eidas.auth.commons.io.SingletonAccessor;
import eu.eidas.auth.engine.configuration.FixedProtocolConfigurationAccessorNoMetadata;
import eu.eidas.auth.engine.configuration.ProtocolConfigurationAccessorNoMetadata;
import eu.eidas.auth.engine.configuration.ProtocolEngineConfigurationNoMetadata;
import eu.eidas.auth.engine.configuration.SamlEngineConfigurationException;
import eu.eidas.auth.engine.configuration.dom.DefaultProtocolEngineConfigurationFactoryNoMetadata;
import eu.eidas.auth.engine.configuration.dom.ProtocolEngineConfigurationFactoryNoMetadata;
import eu.eidas.auth.engine.core.ProtocolProcessorNoMetadataI;
import eu.eidas.engine.exceptions.EIDASSAMLEngineException;
import eu.eidas.samlengineconfig.CertificateConfigurationManager;
import eu.eidas.util.Preconditions;

/**
 * Factory creating {@link ProtocolEngine} instances.
 * <p>
 * Protocol engines are created from a {@link eu.eidas.auth.engine.configuration.ProtocolEngineConfiguration} which are
 * created by a {@link eu.eidas.auth.engine.configuration.dom.ProtocolEngineConfigurationFactory}.
 * <p>
 * One can create one's own ProtocolEngineConfigurationFactory and use it to create one's own ProtocolEngineFactory.
 * <p>
 * There is a default ProtocolEngineFactory: {@link eu.eidas.auth.engine.DefaultProtocolEngineFactory} which uses the
 * default configuration files.
 * <p>
 * One can use the convenient method {@link #getDefaultProtocolEngine(String)} to obtain the default ProtocolEngine
 * created from the default configuration files.
 *
 * @since 1.1
 */
public class ProtocolEngineFactoryNoMetadata {

	private static final Logger LOG = LoggerFactory.getLogger(ProtocolEngineFactoryNoMetadata.class);

	/**
	 * Creates an instance of ProtocolEngine.
	 *
	 * @param nameInstance the name instance
	 * @return instance of ProtocolEngine
	 */
	@Nonnull
	public static ProtocolEngineNoMetadataI createProtocolEngine(@Nonnull String nameInstance,
			@Nonnull ProtocolProcessorNoMetadataI protocolProcessor)
					throws EIDASSAMLEngineException {
		return createProtocolEngine(nameInstance, DefaultProtocolEngineConfigurationFactoryNoMetadata.getInstance(),
				protocolProcessor, new SamlEngineSystemClock());
	}

	@Nonnull
	public static ProtocolEngineNoMetadataI createProtocolEngine(@Nonnull String instanceName,
			@Nonnull CertificateConfigurationManager configManager,
			@Nonnull ProtocolProcessorNoMetadataI protocolProcessor,
			@Nonnull SamlEngineClock samlEngineClock)
					throws SamlEngineConfigurationException {
		ProtocolEngineConfigurationFactoryNoMetadata protocolEngineConfigurationFactory =
				new ProtocolEngineConfigurationFactoryNoMetadata(configManager);

		return createProtocolEngine(instanceName, protocolEngineConfigurationFactory, protocolProcessor,
				samlEngineClock);
	}

	@Nonnull
	public static ProtocolEngineNoMetadataI createProtocolEngine(@Nonnull String instanceName,
			@Nonnull ProtocolEngineConfigurationFactoryNoMetadata protocolEngineConfigurationFactoryNoMetadata,
			@Nonnull ProtocolProcessorNoMetadataI protocolProcessor,
			@Nonnull SamlEngineClock samlEngineClock)
					throws SamlEngineConfigurationException {
		LOG.info(AbstractProtocolEngine.SAML_EXCHANGE, "creating new ProtocolEngine instance: {} ", instanceName);

		ProtocolEngineConfigurationNoMetadata preConfiguration =
				protocolEngineConfigurationFactoryNoMetadata.getConfiguration(instanceName);

		protocolProcessor.configure();

		ProtocolEngineConfigurationNoMetadata configuration = ProtocolEngineConfigurationNoMetadata.builder(preConfiguration)
				.protocolProcessor(protocolProcessor)
				.clock(samlEngineClock)
				.build();

		ProtocolEngineNoMetadataI samlEngine = new ProtocolEngineNoMetadata(new FixedProtocolConfigurationAccessorNoMetadata(configuration));

		LOG.info(AbstractProtocolEngine.SAML_EXCHANGE, "created instance: {} ", samlEngine);

		return samlEngine;
	}

	/**
	 * Returns a default ProtocolEngine instance matching the given name retrieved from the configuration file.
	 *
	 * @param instanceName the instance name
	 * @return the ProtocolEngine instance matching the given name retrieved from the configuration file
	 */
	@Nullable
	public static ProtocolEngineNoMetadataI getDefaultProtocolEngine(@Nonnull String instanceName) {
		Preconditions.checkNotBlank(instanceName, "instanceName");
		return DefaultProtocolEngineFactoryNoMetadata.getInstance().getProtocolEngine(instanceName);
	}

	private final SingletonAccessor<ImmutableMap<String, ProtocolEngineConfigurationNoMetadata>> accessor;

	private final ImmutableMap<String, ProtocolEngineNoMetadataI> engines;

	public ProtocolEngineFactoryNoMetadata(@Nonnull ProtocolEngineConfigurationFactoryNoMetadata configurationFactory)
			throws SamlEngineConfigurationException {
		this(configurationFactory.getConfigurationMapAccessor());
	}

	public ProtocolEngineFactoryNoMetadata(@Nonnull SingletonAccessor<ImmutableMap<String, ProtocolEngineConfigurationNoMetadata>> accessor)
			throws SamlEngineConfigurationException {
		Preconditions.checkNotNull(accessor, "accessor");
		this.accessor = accessor;
		ImmutableMap.Builder<String, ProtocolEngineNoMetadataI> builder = ImmutableMap.builder();
		try {
			for (final String instanceName : accessor.get().keySet()) {
				builder.put(instanceName, createProtocolEngine(instanceName));
			}
		} catch (IOException e) {
			throw new SamlEngineConfigurationException(
					EidasErrors.get(EidasErrorKey.SAML_ENGINE_CONFIGURATION_ERROR.errorCode()),
					EidasErrors.get(EidasErrorKey.SAML_ENGINE_CONFIGURATION_ERROR.errorMessage()), e);
		}
		engines = builder.build();
	}

	@Nonnull
	private ProtocolEngineNoMetadataI createProtocolEngine(@Nonnull final String instanceName) {
		return createProtocolEngine(new ProtocolConfigurationAccessorNoMetadata() {

			@Nonnull
			@Override
			public ProtocolEngineConfigurationNoMetadata get() throws SamlEngineConfigurationException {
				try {
					return accessor.get().get(instanceName);
				} catch (IOException e) {
					throw new SamlEngineConfigurationException(
							EidasErrors.get(EidasErrorKey.SAML_ENGINE_CONFIGURATION_ERROR.errorCode()),
							EidasErrors.get(EidasErrorKey.SAML_ENGINE_CONFIGURATION_ERROR.errorMessage()), e);
				}
			}
		});
	}

	@Nonnull
	protected ProtocolEngineNoMetadataI createProtocolEngine(@Nonnull ProtocolConfigurationAccessorNoMetadata configurationAccessor) {
		return new ProtocolEngineNoMetadata(configurationAccessor);
	}

	/**
	 * Returns a default ProtocolEngine instance matching the given name retrieved from the configuration file.
	 *
	 * @param instanceName the instance name
	 * @return the ProtocolEngine instance matching the given name retrieved from the configuration file
	 */
	@Nullable
	public ProtocolEngineNoMetadataI getProtocolEngine(@Nonnull String instanceName) {
		Preconditions.checkNotBlank(instanceName, "instanceName");
		return engines.get(instanceName.trim());
	}
}
