/*
 * Copyright (c) 2017 by European Commission
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be
 * approved by the European Commission - subsequent versions of the
 * EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/european-union-public-licence-eupl-v.1.1
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the
 * "NOTICE" text file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution.
 * Any derivative works that you distribute must include a readable
 * copy of the "NOTICE" text file.
 */
package eu.eidas.auth.engine;

import java.util.ArrayList;
import java.util.List;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.opensaml.Configuration;
import org.opensaml.common.SignableSAMLObject;
import org.opensaml.saml2.core.Assertion;
import org.opensaml.saml2.core.AuthnRequest;
import org.opensaml.saml2.core.LogoutRequest;
import org.opensaml.saml2.core.LogoutResponse;
import org.opensaml.saml2.core.Response;
import org.opensaml.xml.validation.ValidationException;
import org.opensaml.xml.validation.ValidatorSuite;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;

import eu.eidas.auth.commons.EidasErrorKey;
import eu.eidas.auth.commons.EidasErrors;
import eu.eidas.auth.commons.attribute.AttributeValueTransliterator;
import eu.eidas.auth.commons.protocol.IAuthenticationRequestNoMetadata;
import eu.eidas.auth.commons.protocol.IAuthenticationResponseNoMetadata;
import eu.eidas.auth.commons.protocol.IRequestMessageNoMetadata;
import eu.eidas.auth.commons.protocol.IResponseMessageNoMetadata;
import eu.eidas.auth.commons.protocol.impl.BinaryRequestMessageNoMetadata;
import eu.eidas.auth.commons.protocol.impl.BinaryResponseMessageNoMetadata;
import eu.eidas.auth.engine.configuration.ProtocolConfigurationAccessorNoMetadata;
import eu.eidas.auth.engine.core.eidas.impl.SignCertAliasData;
import eu.eidas.auth.engine.xml.opensaml.CertificateUtil;
import eu.eidas.auth.engine.xml.opensaml.CorrelatedResponse;
import eu.eidas.auth.engine.xml.opensaml.XmlSchemaUtil;
import eu.eidas.engine.exceptions.EIDASSAMLEngineException;

/**
 * The ProtocolEngine is responsible for creating Saml Request and Response from their binary representations and for
 * creating binary representations from Saml Request and Response objects.
 * <p>
 * In eIDAS 1.1, the ProtocolEngine replaces the deprecated SAMLEngine.
 * <p>
 * The protocol engine is responsible for implementing the protocol between the eIDAS Connector and the eIDAS
 * ProxyService.
 * <p>
 * Of course, the default protocol engine strictly implements the eIDAS specification.
 * <p>
 * However the protocol engine can be customized to implement other protocols than eIDAS.
 * <p>
 * A ProtocolEngine instance is obtained from a {@link eu.eidas.auth.engine.ProtocolEngineFactory}.
 * <p>
 * There is a default ProtocolEngineFactory: {@link eu.eidas.auth.engine.DefaultProtocolEngineFactory} which uses the
 * default configuration files.
 * <p>
 * You can obtain the protocol engine named " #MyEngineName# " by using the following statement:
 * <p>
 * {@code ProtocolEngineI protocolEngine = DefaultProtocolEngineFactory.getInstance().getProtocolEngine("#MyEngineName#");}
 * <p>
 * You can also achieve the same result using a convenient method in ProtocolEngineFactory via the
 * getDefaultProtocolEngine method:
 * <p>
 * {@code ProtocolEngineI engine = ProtocolEngineFactory.getDefaultProtocolEngine("#MyEngineName#");}
 *
 * @see ProtocolEngineFactory
 * @see DefaultProtocolEngineFactory
 * @since 1.1
 */
public class ProtocolEngineNoMetadata extends AbstractProtocolEngineNoMetadata implements ProtocolEngineNoMetadataI {

    public static final String ATTRIBUTE_EMPTY_LITERAL = "Attribute name is null or empty.";

    /**
     * The LOG.
     */
    private static final Logger LOG = LoggerFactory.getLogger(ProtocolEngineNoMetadata.class);

    /**
     * Constructs a new Saml engine instance.
     *
     * @param configurationAccessor the accessor to the configuration of this instance.
     */
    public ProtocolEngineNoMetadata(@Nonnull ProtocolConfigurationAccessorNoMetadata configurationAccessor) {
        super(configurationAccessor);
    }

    public static boolean needsTransliteration(String v) {
        return AttributeValueTransliterator.needsTransliteration(v);
    }

    private static void validateSaml2CoreSchema(SignableSAMLObject samlObject) throws EIDASSAMLEngineException {
        LOG.trace("Validate Schema.");
        ValidatorSuite validatorSuite = Configuration.getValidatorSuite("saml2-core-schema-validator");
        try {
            validatorSuite.validate(samlObject);
        } catch (ValidationException e) {
            LOG.info(SAML_EXCHANGE, "BUSINESS EXCEPTION : ValidationException.", e.getMessage());
            LOG.debug(SAML_EXCHANGE, "BUSINESS EXCEPTION : ValidationException.", e);
            throw new EIDASSAMLEngineException(EidasErrors.get(EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode()),
                                               EidasErrors.get(EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorMessage()),
                                               e);
        }
    }

    /**
     * Validate parameters from response.
     *
     * @param request the request
     * @throws EIDASSAMLEngineException the EIDASSAML engine exception
     */
    private void checkRequestSanity(IAuthenticationRequestNoMetadata request) throws EIDASSAMLEngineException {
        getProtocolProcessor().checkRequestSanity(request);
    }

    /**
     * Validate parameters from response.
     *
     * @param response the response authentication request
     * @throws EIDASSAMLEngineException the EIDASSAML engine exception
     */
    @SuppressWarnings("squid:S2583")
    private void checkResponseSanity(IAuthenticationResponseNoMetadata response) throws EIDASSAMLEngineException {
    	boolean allowEmptyRequests = getCoreProperties().allowEmptyRequests();
    	if (!allowEmptyRequests) {
    		if (response.getAttributes() == null || response.getAttributes().isEmpty()) {
    			LOG.error(SAML_EXCHANGE, "No attribute values in response.");
    			throw new EIDASSAMLEngineException(EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
    					EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
    					"No attribute values in response.");
    		}
    	}
    }

    /**
     * Generates the authentication request bytes.
     *
     * @param request the request that contain all parameters for generate an authentication request.
     * @return the EIDAS authentication request that has been processed.
     * @throws EIDASSAMLEngineException the EIDASSAML engine exception
     */
    @Override
    @Nonnull
    public IRequestMessageNoMetadata generateRequestMessage(@Nonnull IAuthenticationRequestNoMetadata request, boolean checkAttrs)
    		throws EIDASSAMLEngineException {
        LOG.trace("Generate SAMLAuthnRequest.");
        if (null == request) {
            LOG.debug(SAML_EXCHANGE, "Sign and Marshall - null input");
            LOG.info(SAML_EXCHANGE, "BUSINESS EXCEPTION : Sign and Marshall -null input");
            throw new EIDASSAMLEngineException(EidasErrorKey.INTERNAL_ERROR.errorCode(),
                                               EidasErrorKey.INTERNAL_ERROR.errorMessage());
        }

        // Validate mandatory parameters
        IAuthenticationRequestNoMetadata requestToBeSent =
                getProtocolProcessor().createProtocolRequestToBeSent(request, getCoreProperties());
        AuthnRequest samlRequest =
                getProtocolProcessor().marshallRequest(requestToBeSent, getCoreProperties(), getClock().getCurrentTime(), checkAttrs);

        try {
            byte[] bytes = signAndMarshallRequest(samlRequest);
            return new BinaryRequestMessageNoMetadata(requestToBeSent, bytes);
        } catch (EIDASSAMLEngineException e) {
            LOG.debug(SAML_EXCHANGE, "Sign and Marshall.", e);
            LOG.info(SAML_EXCHANGE, "BUSINESS EXCEPTION : Sign and Marshall.", e);
            throw new EIDASSAMLEngineException(EidasErrorKey.INTERNAL_ERROR.errorCode(),
                                               EidasErrorKey.INTERNAL_ERROR.errorMessage(), e);
        }
    }

    /**
     * Generate authentication response in one of the supported formats.
     *
     * @param request the request
     * @param response the authentication response from the IdP
     * @param ipAddress the IP address
     * @param signAssertion whether to sign the attribute assertion
     * @return the authentication response
     * @throws EIDASSAMLEngineException the EIDASSAML engine exception
     */
    @Override
    @Nonnull
    public IResponseMessageNoMetadata generateResponseMessage(@Nonnull IAuthenticationRequestNoMetadata request,
                                                    @Nonnull IAuthenticationResponseNoMetadata response,
                                                    boolean signAssertion,
                                                    @Nonnull String ipAddress) throws EIDASSAMLEngineException {
        LOG.trace("generateResponseMessage");
        // Validate parameters
        validateParamResponse(request, response);

        Response samlResponse =
                getProtocolProcessor().marshallResponse(request, response, ipAddress, getCoreProperties(), getClock().getCurrentTime());

        // update the assertions in the response to signed assertions if needed:
        if (signAssertion) {
            List<Assertion> assertions = samlResponse.getAssertions();
            List<Assertion> signedAssertions = new ArrayList<>(assertions.size());
            for (Assertion assertion : assertions) {
                try {
                    Assertion signedAssertion = signAssertion(assertion);
                    signedAssertions.add(signedAssertion);
                } catch (EIDASSAMLEngineException e) {
                    LOG.error(SAML_EXCHANGE, "BUSINESS EXCEPTION : cannot sign assertion: " + e, e);
                    throw e;
                }
            }
            samlResponse.getAssertions().clear();
            samlResponse.getAssertions().addAll(signedAssertions);
        }
        return encryptAndSignAndMarshallResponse(request, response, samlResponse);
    }

    private IResponseMessageNoMetadata encryptAndSignAndMarshallResponse(@Nonnull IAuthenticationRequestNoMetadata request,
                                                               @Nonnull IAuthenticationResponseNoMetadata response,
                                                               Response samlResponse) throws EIDASSAMLEngineException {
        // encrypt and sign the whole response:
        try {
            byte[] responseBytes = signAndMarshallResponse(request, samlResponse);
            return new BinaryResponseMessageNoMetadata(response, responseBytes);
        } catch (EIDASSAMLEngineException e) {
            LOG.error(SAML_EXCHANGE, "BUSINESS EXCEPTION : Sign and Marshall: " + e, e);
            throw new EIDASSAMLEngineException(EidasErrorKey.INTERNAL_ERROR.errorCode(),
                                               EidasErrorKey.INTERNAL_ERROR.errorMessage(), e);
        }
    }

    /**
     * Generates an authentication response error message.
     *
     * @param request the request
     * @param response the response
     * @param ipAddress the IP address
     * @return the authentication response
     * @throws EIDASSAMLEngineException the EIDASSAML engine exception
     */
    @Override
    @Nonnull
    public IResponseMessageNoMetadata generateResponseErrorMessage(@Nonnull IAuthenticationRequestNoMetadata request,
                                                         @Nonnull IAuthenticationResponseNoMetadata response,
                                                         @Nonnull String ipAddress) throws EIDASSAMLEngineException {

        Response responseFail =
                getProtocolProcessor().marshallErrorResponse(request, response, ipAddress, getCoreProperties(), getClock().getCurrentTime());

        IAuthenticationResponseNoMetadata authenticationResponse =
                getProtocolProcessor().unmarshallErrorResponse(response, responseFail, ipAddress, getCoreProperties());

        LOG.trace("Sign and Marshall ResponseFail.");
        return encryptAndSignAndMarshallResponse(request, authenticationResponse, responseFail);
    }

    /**
     * Unmarshalls the given bytes into a SAML Request.
     *
     * @param requestBytes the SAML request bytes
     * @return the SAML request instance
     * @throws EIDASSAMLEngineException the EIDASSAML engine exception
     */
    public SignCertAliasData unmarshallRequest(@Nonnull byte[] requestBytes) throws EIDASSAMLEngineException {
        LOG.trace("Validate request bytes.");

        if (null == requestBytes) {
            LOG.info(SAML_EXCHANGE, "BUSINESS EXCEPTION : Saml request bytes are null.");
            throw new EIDASSAMLEngineException(EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
                                               EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
                                               "Saml request bytes are null.");
        }

        Document document = XmlSchemaUtil.validateSamlSchema(requestBytes);
        AuthnRequest request = (AuthnRequest) unmarshall(document);
        
        SignCertAliasData ret = validateSignature(request);
        
        request = (AuthnRequest) ret.getSignature();

        validateSaml2CoreSchema(request);

        validateRequestWithValidatorSuite(request);

        return ret;
    }

    /**
     * Process and validates the authentication request.
     *
     * @param requestBytes the token SAML
     * @return the authentication request
     * @throws EIDASSAMLEngineException the EIDASSAML engine exception
     */
    @Override
    @Nonnull
    public IAuthenticationRequestNoMetadata unmarshallRequestAndValidate(@Nonnull byte[] requestBytes)
            throws EIDASSAMLEngineException {
        LOG.trace("processValidateRequestToken");

        if (null == requestBytes) {
            LOG.info(SAML_EXCHANGE, "BUSINESS EXCEPTION : Saml authentication request is null.");
            throw new EIDASSAMLEngineException(EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
                                               EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
                                               "Saml authentication request is null.");
        }
        SignCertAliasData ret = unmarshallRequest(requestBytes);
        AuthnRequest originalSamlRequest = (AuthnRequest) ret.getSignature();
        LOG.trace("Generate EIDASAuthnSamlRequest.");

        String originCountryCode = (originalSamlRequest.getSignature() != null) ? CertificateUtil.getCountry(
                originalSamlRequest.getSignature().getKeyInfo()) : null;

        IAuthenticationRequestNoMetadata authenticationRequest =
                getProtocolProcessor().unmarshallRequest(originalSamlRequest,
                		originCountryCode, ret.getAlias(), ret.getCertificate());

        checkRequestSanity(authenticationRequest);

        return authenticationRequest;
    }

    /**
     * Unmarshalls the given bytes into a SAML Response.
     *
     * @param tokenSaml the SAML response bytes
     * @return the SAML response instance
     * @throws EIDASSAMLEngineException the EIDASSAML engine exception
     */
    @Override
    @Nonnull
    public Correlated unmarshallResponse(@Nonnull byte[] responseBytes) throws EIDASSAMLEngineException {
        LOG.trace("Validate response bytes.");

        if (null == responseBytes) {
            LOG.info(SAML_EXCHANGE, "BUSINESS EXCEPTION : Saml response bytes are null.");
            throw new EIDASSAMLEngineException(EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
                                               EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
                                               "Saml response bytes are null.");
        }

        LOG.trace("Generate SAML Response.");

        Document document = XmlSchemaUtil.validateSamlSchema(responseBytes);
        Response response = (Response) unmarshall(document);
        response = validateSignatureAndDecryptAndValidateAssertionSignatures(response);
        
        validateSaml2CoreSchema(response);

        validateResponseWithValidatorSuite(response);

        return new CorrelatedResponse(response);
    }

    /**
     * Process and validates the authentication response.
     *
     * @param responseBytes the token SAML
     * @param userIpAddress the user IP
     * @return the authentication response
     * @throws EIDASSAMLEngineException the EIDASSAML engine exception
     */
    @Override
    @Nonnull
    public IAuthenticationResponseNoMetadata unmarshallResponseAndValidate(@Nonnull byte[] responseBytes,
                                                                 @Nonnull String userIpAddress,
                                                                 long beforeSkewTimeInMillis,
                                                                 long afterSkewTimeInMillis,
                                                                 @Nullable String audienceRestriction)
            throws EIDASSAMLEngineException {
        Correlated samlResponse = unmarshallResponse(responseBytes);

        return validateUnmarshalledResponse(samlResponse, userIpAddress, beforeSkewTimeInMillis, afterSkewTimeInMillis, audienceRestriction);
    }

    private void validateAssertionSignatures(Response response) throws EIDASSAMLEngineException {
        try {
            boolean validateSign = getCoreProperties().isValidateSignature();
            if (validateSign) {
                for (Assertion assertion : response.getAssertions()) {
                    if (assertion.isSigned() && null != assertion.getSignature()) {
                        getSigner().validateSignature(assertion, null);
                    }
                }
            }
        } catch (EIDASSAMLEngineException e) {
            EIDASSAMLEngineException exc =
                    new EIDASSAMLEngineException(EidasErrorKey.INVALID_ASSERTION_SIGNATURE.errorCode(),
                                                 EidasErrorKey.INVALID_ASSERTION_SIGNATURE.errorMessage(), e);
            throw exc;
        }
    }

    /**
     * Validate parameters from response.
     *
     * @param request the request
     * @param response the response authentication request
     * @throws EIDASSAMLEngineException the EIDASSAML engine exception
     */
    private void validateParamResponse(IAuthenticationRequestNoMetadata request, IAuthenticationResponseNoMetadata response)
            throws EIDASSAMLEngineException {
        LOG.trace("Validate parameters response.");
        checkRequestSanity(request);
        checkResponseSanity(response);
    }

    private void validateRequestWithValidatorSuite(@Nonnull SignableSAMLObject request) throws EIDASSAMLEngineException {
        String suiteId = getProtocolProcessor().getRequestValidatorId();
        LOG.trace("Validate AuthnRequest with suite: " + suiteId);
        ValidatorSuite suite = Configuration.getValidatorSuite(suiteId);
        try {
            suite.validate(request);
            LOG.debug("AuthnRequest validation with " + suiteId + " succeeded.");
        } catch (ValidationException e) {
            LOG.error(SAML_EXCHANGE, "BUSINESS EXCEPTION : validate AuthnRequest: " + e, e);
            throw new EIDASSAMLEngineException(EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
                                               EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorMessage(), e);
        }
    }

    private void validateResponseWithValidatorSuite(@Nonnull Response response) throws EIDASSAMLEngineException {
        String suiteId = getProtocolProcessor().getResponseValidatorId();
        LOG.trace("Validate Response with suite: " + suiteId);
        ValidatorSuite suite = Configuration.getValidatorSuite(suiteId);
        try {
            suite.validate(response);
            LOG.debug("Response validation with " + suiteId + " succeeded.");
        } catch (ValidationException e) {
            LOG.error(SAML_EXCHANGE, "BUSINESS EXCEPTION : validate Response: " + e, e);
            throw new EIDASSAMLEngineException(EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
                                               EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorMessage(), e);
        }
    }

    private SignCertAliasData validateSignature(SignableSAMLObject request) throws EIDASSAMLEngineException {
        boolean validateSign = getCoreProperties().isValidateSignature();
        if (validateSign) {
            LOG.trace("Validate request Signature.");
            try {
                return getSigner().validateSignatureWithData(request, null);
            } catch (EIDASSAMLEngineException e) {
                LOG.error(SAML_EXCHANGE, "BUSINESS EXCEPTION : SAMLEngineException validateSignature: " + e,
                          e.getMessage(), e);
                throw e;
            }
        } else {
        	SignCertAliasData ret = new SignCertAliasData();
        	ret.setSignature(request);
        	return ret;
        }
    }

    private Response validateSignatureAndDecryptAndValidateAssertionSignatures(Response response)
            throws EIDASSAMLEngineException {
        Response validResponse = response;
        LOG.trace("Validate response Signature.");
        if (!validResponse.isSigned() || null == validResponse.getSignature()) {
            throw new EIDASSAMLEngineException(EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(), "No signature");
        }

        String country = CertificateUtil.getCountry(validResponse.getSignature().getKeyInfo());
        LOG.debug(SAML_EXCHANGE, "Response received from country: " + country);
        try {
            validResponse = validateSignatureAndDecrypt(validResponse);

            validateAssertionSignatures(validResponse);
        } catch (EIDASSAMLEngineException e) {
            LOG.error(SAML_EXCHANGE, "BUSINESS EXCEPTION : SAMLEngineException validateSignature: " + e,
                      e.getMessage(), e);
            throw e;
        } 
        return validResponse;
    }
    
    private LogoutResponse validateLogoutSignature(LogoutResponse response) throws EIDASSAMLEngineException {
        boolean validateSign = getCoreProperties().isValidateSignature();
        if (validateSign) {
            LOG.trace("Validate response Signature.");
            if (!response.isSigned() || null == response.getSignature()) {
                throw new EIDASSAMLEngineException(EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(), "No signature");
            }

            try {
            	getSigner().validateSignatureWithData(response, null);
            } catch (EIDASSAMLEngineException e) {
                LOG.error(SAML_EXCHANGE, "BUSINESS EXCEPTION : SAMLEngineException validateSignature: " + e,
                          e.getMessage(), e);
                throw e;
            }
        }
        return response;
    }

    /**
     * Validate authentication response.
     *
     * @param unmarshalledResponse the token SAML
     * @param userIpAddress the user IP
     * @param beforeSkewTimeInMillis the skew time for notBefore (value to be added)
     * @param afterSkewTimeInMillis the skew time for notOnOrAfter (value to be added)
     * @return the authentication response
     * @throws EIDASSAMLEngineException the EIDASSAML engine exception
     */
    @Nonnull
    @Override
    public IAuthenticationResponseNoMetadata validateUnmarshalledResponse(@Nonnull Correlated unmarshalledResponse,
                                                                @Nonnull String userIpAddress,
                                                                long beforeSkewTimeInMillis,
                                                                long afterSkewTimeInMillis,
                                                                @Nullable String audienceRestriction)
            throws EIDASSAMLEngineException {

        Response response = ((CorrelatedResponse) unmarshalledResponse).getResponse();

        return getProtocolProcessor().unmarshallResponse(response, getCoreProperties().isIpValidation(), userIpAddress,
                beforeSkewTimeInMillis, afterSkewTimeInMillis, getClock().getCurrentTime(),
                                                         audienceRestriction);
    }

	@Override
	public byte[] generateLogoutRequestMessage(String assertionConsumerUrl, String providerName, 
			String destination, String samlId)
			throws EIDASSAMLEngineException {

        LogoutRequest samlRequest =
                getProtocolProcessor().marshallLogoutRequest(assertionConsumerUrl, providerName,
                		destination, samlId, getCoreProperties(), getClock().getCurrentTime());

        try {
        	return marshall(getSigner().sign(samlRequest));
        } catch (EIDASSAMLEngineException e) {
            LOG.debug(SAML_EXCHANGE, "Sign and Marshall.", e);
            LOG.info(SAML_EXCHANGE, "BUSINESS EXCEPTION : Sign and Marshall.", e);
            throw new EIDASSAMLEngineException(EidasErrorKey.INTERNAL_ERROR.errorCode(),
                                               EidasErrorKey.INTERNAL_ERROR.errorMessage(), e);
        }
	}
	
	@Override
	public byte[] generateLogoutResponseMessage(@Nonnull final String issuer, 
    		@Nonnull final String samlId, @Nonnull final String destination)
			throws EIDASSAMLEngineException {
		LOG.trace("generateResponseMessage");
        LogoutResponse samlResponse =
                getProtocolProcessor().marshallLogoutResponse(issuer, samlId, destination,
                		getCoreProperties(), getClock().getCurrentTime());
        
        return marshall(getSigner().sign(samlResponse));
    }
	

	@Override
	public LogoutRequest unmarshallLogoutRequestAndValidate(byte[] requestBytes)
			throws EIDASSAMLEngineException {
		LOG.trace("Validate request bytes.");

        if (null == requestBytes) {
            LOG.info(SAML_EXCHANGE, "BUSINESS EXCEPTION : Saml request bytes are null.");
            throw new EIDASSAMLEngineException(EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
                                               EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
                                               "Saml request bytes are null.");
        }

        Document document = XmlSchemaUtil.validateSamlSchema(requestBytes);
        LogoutRequest request = (LogoutRequest) unmarshall(document);
        
        SignCertAliasData ret = validateSignature(request);
        
        //ret.getAlias(), ret.getCertificate()
        
        request = (LogoutRequest) ret.getSignature();

        validateSaml2CoreSchema(request);

        validateRequestWithValidatorSuite(request);

        LogoutRequest originalSamlRequest = (LogoutRequest) ret.getSignature();
        LOG.trace("Generate EIDASAuthnSamlRequest.");

        LogoutRequest authenticationRequest =
                getProtocolProcessor().unmarshallLogoutRequest(originalSamlRequest);

        return authenticationRequest;
	}

	@Override
	public LogoutResponse unmarshallLogoutResponseAndValidate(byte[] responseBytes, String userIpAddress,
			long beforeSkewTimeInMillis, long afterSkewTimeInMillis, String audienceRestriction)
			throws EIDASSAMLEngineException {
		LOG.trace("Validate response bytes.");

        if (null == responseBytes) {
            LOG.info(SAML_EXCHANGE, "BUSINESS EXCEPTION : Saml response bytes are null.");
            throw new EIDASSAMLEngineException(EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
                                               EidasErrorKey.MESSAGE_VALIDATION_ERROR.errorCode(),
                                               "Saml response bytes are null.");
        }

        LOG.trace("Generate SAML Response.");

        Document document = XmlSchemaUtil.validateSamlSchema(responseBytes);
        LogoutResponse response = (LogoutResponse) unmarshall(document);
        response = validateLogoutSignature(response);
        
        validateSaml2CoreSchema(response);

        return getProtocolProcessor().unmarshallLogoutResponse(response, getCoreProperties().isIpValidation(), userIpAddress,
                beforeSkewTimeInMillis, afterSkewTimeInMillis, getClock().getCurrentTime(),
                                                         audienceRestriction);
	}
}
