package eu.eidas.auth.engine;

import java.security.cert.X509Certificate;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.opensaml.saml2.core.LogoutRequest;
import org.opensaml.saml2.core.LogoutResponse;

import eu.eidas.auth.commons.protocol.IAuthenticationRequestNoMetadata;
import eu.eidas.auth.commons.protocol.IAuthenticationResponseNoMetadata;
import eu.eidas.auth.commons.protocol.IRequestMessageNoMetadata;
import eu.eidas.auth.commons.protocol.IResponseMessageNoMetadata;
import eu.eidas.auth.commons.tx.CorrelationMap;
import eu.eidas.auth.engine.core.ProtocolCipherI;
import eu.eidas.auth.engine.core.ProtocolProcessorNoMetadataI;
import eu.eidas.auth.engine.core.ProtocolSignerI;
import eu.eidas.auth.engine.core.SamlEngineCoreProperties;
import eu.eidas.engine.exceptions.EIDASSAMLEngineException;

/**
 * Engine providing the implementation of the actual protocol (e.g. SAML, eIDAS, etc).
 *
 * @since 1.1
 */
public interface ProtocolEngineNoMetadataI {

    /**
     * Generates the bytes of the request based on the given request data for the given target service.
     *
     * @param request the request data
     * @param checkAttrs check attrs schema
     * @return the container of the request bytes
     * @throws EIDASSAMLEngineException
     */
    @Nonnull
    IRequestMessageNoMetadata generateRequestMessage(@Nonnull IAuthenticationRequestNoMetadata request,
    		@Nonnull final boolean checkAttrs)
            throws EIDASSAMLEngineException;

    @Nonnull
    IResponseMessageNoMetadata generateResponseMessage(@Nonnull IAuthenticationRequestNoMetadata request,
                                             @Nonnull IAuthenticationResponseNoMetadata response,
                                             boolean signAssertion,
                                             @Nonnull String ipAddress) throws EIDASSAMLEngineException;

    /**
     * Generates the bytes of the request based on the given request data for the given target service.
     *
     * @param request the request data
     * @param serviceIssuer the target service to which the request is to be sent
     * @return the container of the request bytes
     * @throws EIDASSAMLEngineException
     */
    @Nonnull
    byte[] generateLogoutRequestMessage(@Nonnull final String assertionConsumerUrl, @Nonnull final String providerName, 
    		@Nonnull final String destination, @Nonnull final String samlId)
            throws EIDASSAMLEngineException;
    
    @Nonnull
    byte[] generateLogoutResponseMessage(@Nonnull final String issuer, 
    		@Nonnull final String samlId, @Nonnull final String destination) throws EIDASSAMLEngineException;
    
    @Nonnull
    IResponseMessageNoMetadata generateResponseErrorMessage(@Nonnull IAuthenticationRequestNoMetadata request,
                                                  @Nonnull IAuthenticationResponseNoMetadata response,
                                                  @Nonnull String ipAddress) throws EIDASSAMLEngineException;

    @Nullable
    ProtocolCipherI getCipher();

    @Nonnull
    SamlEngineClock getClock();

    @Nonnull
    SamlEngineCoreProperties getCoreProperties();

    @Nullable
    X509Certificate getDecryptionCertificate() throws EIDASSAMLEngineException;

    @Nonnull
    ProtocolProcessorNoMetadataI getProtocolProcessor();

    @Nonnull
    String getInstanceName();

    @Nonnull
    ProtocolSignerI getSigner();

    @Nonnull
    X509Certificate getSigningCertificate() throws EIDASSAMLEngineException;

    @Nonnull
    IAuthenticationRequestNoMetadata unmarshallRequestAndValidate(@Nonnull byte[] requestBytes)
            throws EIDASSAMLEngineException;

    @Nonnull
    LogoutRequest unmarshallLogoutRequestAndValidate(@Nonnull byte[] requestBytes)
            throws EIDASSAMLEngineException;
    
    /**
     * Unmarshalls the given bytes into a {@link Correlated} response object.
     * <p>
     * The {@link Correlated} response object is used to retrieve the correlated request object in a {@link
     * CorrelationMap} before invoking {@link #validateUnmarshalledResponse(Correlated, String, long)} to obtain the
     * complete response.
     *
     * @param tokenSaml the SAML response bytes
     * @return the SAML response instance
     * @throws EIDASSAMLEngineException the EIDASSAML engine exception
     */
    @Nonnull
    Correlated unmarshallResponse(@Nonnull byte[] responseBytes) throws EIDASSAMLEngineException;

    @Nonnull
    IAuthenticationResponseNoMetadata unmarshallResponseAndValidate(@Nonnull byte[] responseBytes,
                                                          @Nonnull String userIpAddress,
                                                          long beforeSkewTimeInMillis,
                                                          long afterSkewTimeInMillis,
                                                          @Nullable String audienceRestriction)
            throws EIDASSAMLEngineException;
    
    @Nonnull
    LogoutResponse unmarshallLogoutResponseAndValidate(@Nonnull byte[] responseBytes,
                                                          @Nonnull String userIpAddress,
                                                          long beforeSkewTimeInMillis,
                                                          long afterSkewTimeInMillis,
                                                          @Nullable String audienceRestriction)
            throws EIDASSAMLEngineException;

    @Nonnull
    IAuthenticationResponseNoMetadata validateUnmarshalledResponse(@Nonnull Correlated unmarshalledResponse,
                                                         @Nonnull String userIpAddress,
                                                         long beforeSkewTimeInMillis,
                                                         long afterSkewTimeInMillis,
                                                         @Nullable String audienceRestriction)
            throws EIDASSAMLEngineException;
}
