package eu.eidas.auth.engine.configuration.dom;

import java.io.IOException;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.ImmutableMap;

import eu.eidas.auth.commons.io.SingletonAccessor;
import eu.eidas.auth.commons.lang.reflect.ReflectionUtil;
import eu.eidas.auth.engine.SamlEngineClock;
import eu.eidas.auth.engine.configuration.ProtocolEngineConfigurationNoMetadata;
import eu.eidas.auth.engine.core.ProtocolCipherI;
import eu.eidas.auth.engine.core.ProtocolProcessorNoMetadataI;
import eu.eidas.auth.engine.core.ProtocolSignerI;
import eu.eidas.auth.engine.core.SamlEngineCoreProperties;

/**
 * ReloadableConfiguration InvocationHandler
 *
 * @since 1.1
 */
@VisibleForTesting
public final class ReloadableProtocolConfigurationInvocationHandlerNoMetadata<T> implements InvocationHandler {

	/**
	 * The logger.
	 */
	private static final Logger LOG = LoggerFactory.getLogger(ReloadableProtocolConfigurationInvocationHandlerNoMetadata.class);

	@Nonnull
	private final String name;

	@Nonnull
	private final SingletonAccessor<ImmutableMap<String, ProtocolEngineConfigurationNoMetadata>> fileAccessor;

	@Nonnull
	private final ConfigurationGetter<T> getter;

	interface ConfigurationGetter<T> {

		T get(@Nonnull ProtocolEngineConfigurationNoMetadata configuration);
	}

	ReloadableProtocolConfigurationInvocationHandlerNoMetadata(@Nonnull final String nameVal,
			@Nonnull
			final SingletonAccessor<ImmutableMap<String, ProtocolEngineConfigurationNoMetadata>> fileAccessorVal,
			@Nonnull final ConfigurationGetter<T> getterVal) {
		name = nameVal;
		fileAccessor = fileAccessorVal;
		getter = getterVal;
	}

	@Nonnull
	private static ProtocolEngineConfigurationNoMetadata getNamedConfiguration(@Nonnull String name,
			@Nonnull
			SingletonAccessor<ImmutableMap<String, ProtocolEngineConfigurationNoMetadata>> accessor) {
		ProtocolEngineConfigurationNoMetadata configuration;
		try {
			configuration = accessor.get().get(name);
		} catch (IOException e) {
			throw new IllegalStateException("Configuration instance \"" + name + "\" could not be loaded: " + e, e);
		}
		if (null == configuration) {
			throw new IllegalStateException("Configuration instance \"" + name + "\" does not exist.");
		}
		return configuration;
	}

	@Nonnull
	static ProtocolEngineConfigurationNoMetadata newConfigurationProxy(@Nonnull String name,
			@Nonnull
			SingletonAccessor<ImmutableMap<String, ProtocolEngineConfigurationNoMetadata>> accessor) {
		SamlEngineCoreProperties samlEngineCoreProperties =
				newProxyInstance(SamlEngineCoreProperties.class, name, accessor,
						new ConfigurationGetter<SamlEngineCoreProperties>() {

					@Override
					public SamlEngineCoreProperties get(
							@Nonnull ProtocolEngineConfigurationNoMetadata configuration) {
						return configuration.getCoreProperties();
					}
				});
		ProtocolSignerI signer =
				newProxyInstance(ProtocolSignerI.class, name, accessor, new ConfigurationGetter<ProtocolSignerI>() {

					@Override
					public ProtocolSignerI get(@Nonnull ProtocolEngineConfigurationNoMetadata configuration) {
						return configuration.getSigner();
					}
				});
		ProtocolCipherI cipher =
				newProxyInstance(ProtocolCipherI.class, name, accessor, new ConfigurationGetter<ProtocolCipherI>() {

					@Override
					public ProtocolCipherI get(@Nonnull ProtocolEngineConfigurationNoMetadata configuration) {
						return configuration.getCipher();
					}
				});
		ProtocolProcessorNoMetadataI protocolProcessor = newProxyInstance(ProtocolProcessorNoMetadataI.class, name, accessor,
				new ConfigurationGetter<ProtocolProcessorNoMetadataI>() {

			@Override
			public ProtocolProcessorNoMetadataI get(@Nonnull
					ProtocolEngineConfigurationNoMetadata configuration) {
				return configuration.getProtocolProcessor();
			}
		});
		SamlEngineClock clock =
				newProxyInstance(SamlEngineClock.class, name, accessor, new ConfigurationGetter<SamlEngineClock>() {

					@Override
					public SamlEngineClock get(@Nonnull ProtocolEngineConfigurationNoMetadata configuration) {
						return configuration.getClock();
					}
				});

		return ProtocolEngineConfigurationNoMetadata.builder()
				.instanceName(name)
				.coreProperties(samlEngineCoreProperties)
				.signer(signer)
				.cipher(cipher)
				.protocolProcessor(protocolProcessor)
				.clock(clock)
				.build();
	}

	@Nullable
	@SuppressWarnings("unchecked")
	private static <T> T newProxyInstance(@Nonnull Class<T> type,
			@Nonnull String name,
			@Nonnull
			SingletonAccessor<ImmutableMap<String, ProtocolEngineConfigurationNoMetadata>> accessor,
			@Nonnull ConfigurationGetter<T> getter) {
		ProtocolEngineConfigurationNoMetadata configuration = getNamedConfiguration(name, accessor);
		T proxiedObject = getter.get(configuration);
		if (null == proxiedObject) {
			return null;
		}
		ReloadableProtocolConfigurationInvocationHandlerNoMetadata<T> invocationHandler =
				new ReloadableProtocolConfigurationInvocationHandlerNoMetadata<>(name, accessor, getter);

		return ReflectionUtil.newProxyInstance(ReloadableConfigurationMap.class.getClassLoader(), type,
				(Class<? extends T>) proxiedObject.getClass(), invocationHandler);
	}

	public T getProxiedObject() {
		ProtocolEngineConfigurationNoMetadata configuration = getNamedConfiguration(name, fileAccessor);
		return getter.get(configuration);
	}

	@Override
	public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
		try {
			T instance = getProxiedObject();
			if (null == instance) {
				// Houston we have a problem!
				// The configuration has been modified and is now allowing a null configuration property (e.g. null cipher)
				// TODO to fix this, we should not allow a null cipher but instead we should have a Cipher object doing nothing
				return null;
			}
			return method.invoke(instance, args);
		} catch (InvocationTargetException e) {
			LOG.error("", e);
			throw e.getTargetException();
		}
	}
}
